"""
Key structures in pySubnetSB

A NamedMatrix is a two dimensional numpy array with row and column names.
A Network is a representation of a chemical reaction network that includes a NamedMatrix for reactants and products.
A Constraint describes numerical or categorical charateristics of a network that are organized as one or more NamedMatrix.
"""
__version__ = "1.0.2"
